⍝ Binomial coefficent algorithm that attempts to maximize the integer
⍝ range of its result.  The GNU APL primitive ! yields results in
⍝ floating point representation thereby losing least significant
⍝ digits when it uses the factorial primitive on integers greater than
⍝ 20.

⍝ This implementation is due to Jim Weigang.

∇ Z ← A BINOM B; PRIMES; TA; TB
  ⍝ Computes A ! B as ( ! B ) ÷ ( ! A ) × ! B − A
  Z ← 0
  → ( ( A ≥ 0 ) ∧ ( B ≥ 0 ) ∧ B < A ) / 0
  → ( ( A < 0 ) ∧ ( B ≥ 0 ) ∧ B ≥ A ) / 0
  → ( ( A < 0 ) ∧ ( B < 0 ) ∧ B < A ) / 0
  → ( ∼ ( A ≥ 0 ) ∧ ( B ≥ 0 ) ∧ B ≥ A ) / L1
  PRIMES ← PRIMESTO B
  Z ← NFP ( PFNF B ) - ( PFNF A ) + PFNF B - A
  → 0
 L1: → ( ∼ ( A ≥ 0 ) ∧ B < 0 ) / L2
  TB ← A - B + 1
  PRIMES ← PRIMESTO TB
  Z ← ( ¯1 ⋆ A ) × NFP ( PFNF TB ) − ( PFNF A ) + PFNF TB - A
  → 0
  ⍝ ( A < 0 ) ∧ ( B < 0 ) ∧ B < A
 L2: TA ← − B + 1
  TB ← | A + 1
  PRIMES ← PRIMESTO TB
  Z ← ( ¯1 ⋆ B - A ) × NFP ( PFNF TB ) − ( PFNF TA ) + PFNF TB - TA
∇

∇ P ← PFNF N; M
  ⍝ Computes prime power representation of ! ⍵
  ⍝ M ← ×\ ⍉ ( ( ⌊ 2 ⍟ 1 ⌈ N ), ⍴ PRIMES ) ⍴ PRIMES
  P ← ( ⍴ PRIMES ) ⍴ 0
  → ( N = 1 ) / 0
  M ← ×\ ⍉ 1 ⍪ ( ( ⌊ 2 ⍟ 1 ⌈ N ), ⍴ PRIMES ) ⍴ PRIMES
  M ← ( ( ( ¯1 + ⍴ ⍴ M ) ⍴ 0 ) , 1 ) ↓ M
  P ← +/ ⌊ N ÷ M
∇

∇ N ← NFP P
  ⍝ Converts prime-powers vector ⍵ back to a number
  N ← ×/ P / PRIMES
∇

∇ P ← PFN N
  ⍝ Converts number ⍵ to a prime-powers vector
  P ← PRIMES FRCOUNT ALLFACTORS N
∇

∇ Z ← PRIMESTO N; B; P; ⎕IO
  ⍝ Returns primes up to ⍵
  Z ← ⍳ ⎕IO ← 0
  N ← N + 1
  B ← N ↑ 1 1
 L1: P ← B ⍳ 0
  → ( P = N ) / 0
  Z ← Z , P
  B ← B ∨ N ⍴ P ↑ 1
  → L1
∇

∇ Z ← P ALLFACTORS A; B; I; J; N
  ⍝ Returns the prime factors of the numbers in ⍵
  → ( 2 = ⎕NC 'P' ) / L1
  ⍝ If primes list wasn't provided, generate it
  P ← PRIMESTO ⌊ ( 0 ⌈ ⌈/ A ) ⋆ .5
 L1: Z ← 0 ⍴ A ← , A
  I ← 1 ⋄ N ← ⍴ P
  ⍝ Dive on in.
  → L3
  ⍝ Loop for each prime.
 L2: → ( N < I ← I + 1 ) / L4
  ⍝ Indices of As that are divisible by P[ I ].
  B ← 0 = P[ I ] | A
  ⍝ On to next prime if none
  → ( ~ 1 ∊ B ) / L2
  ⍝ Indices of divisible As.
  J ← B / ⍳ ⍴ B
  ⍝ Add the P[ I ]s to the result.
  Z ← Z , ( ⍴ J ) ⍴ P[ I ]
  ⍝ Remove the factors.
  A[ J ] ← A[ J ] ÷ P[ I ]
  ⍝ Append As which are prime.
 L3: Z ← Z , ( B ← A ∊ P ) / A
  ⍝ And remove them from further consideration.
  A ← ( ~ B ) / A
  ⍝ Index of largest needed prime.
  N ← ¯1 + ( P ≤ ⌊ ( 0 ⌈ ⌈/ A ) ⋆ .5 ) ⍳ 0
  ⍝ Try same prime again.
  I ← I − 1
  ⍝ End loop.
  → L2
  ⍝ Leftovers are prime, too.
 L4: Z ← Z , A
  ⍝ Standardize the order.
  Z ← Z[ ⍋ Z ]
∇

∇ Z ← D FRCOUNT V; B; F; I
  ⍝ Counts frequency of values ⍺ in sorted vector ⍵.
  ⍝ First marks.
  B ← V ≠ ¯1 ⌽ V ⋄ B[ ⍳ × ⍴ B ] ← 1
  ⍝ Indices of firsts.
  I ← B / ⍳ ⍴ B
  ⍝ Frequency of each kind.
  F ← ( 1 ↓ I , 1 + ⍴ B ) - I
  Z ← ( ⍴ D ) ⍴ 0
  ⍝ Put them where they belong.
  Z[ D ⍳ V[ I ] ] ← F
∇

⍝???
∇ Z ← P FACTORNF N; M
  ⍝ Returns the prime factors of ! ⍵
  ⍝ From Knuth, The Art of Comp. Prog., Vol. 1, P. 46
  → ( 2 = ⎕NC 'P' ) / L1
  ⍝ Compute primes if caller didn't provide them.
  P ← PRIMESTO N
 L1: M ← × \ ⍉ ⌊ 2 ⍟ 1 ⌈ N ) , ⍴ P ) ⍴ P
  Z ← +/ ( ⌊ N ÷ M ) / P
∇

⍝ Tests
⎕ ← PRIMES ← PRIMESTO 13

⎕ ← N ← 115615500

⍝ Prime powers representation: 2 1 3 2 2 1
⎕ ← P ← PFN N

⍝ Convert back to N: 115615500
NFP P

⍝ Prime powers of ! 12: 10 5 2 1 1 0
PFNF 12

NFP PFNF 12

! 12

10 BINOM 21

10 ! 21

∇ NAME ALT EXP; J
  J ← ⎕FX ⊃ ( ⊂ 'Z ← ' , NAME ) , ⊂ 'Z ← ' , EXP
  ⍝ INTRANSITIVE SYNONYM
∇

∇Z ← tr TIME F; T0; T1; I
 ⍝ Timer for function runs.  Input F is a char. vector, consisting of
 ⍝ the function name and arguments, separated by blanks as usual.  Out-
 ⍝ put is the time in milliseconds for a run, averaged over tr runs.
 ⍝ If no tr is given, the default is assumed.  The function value from
 ⍝ the last run is saved in the global variable XF .
 ⍝ Default the TIME repetition
 ⍎ ( 0 = ⎕NC 'tr' ) / 'tr ← 1'
 I ← 1
 Z ← ⍳ 0
 'f' ALT F
 REK: T0 ← ( 100 60 60 1000 ) ⊥ ¯4 ↑ ⎕TS
 ⍝ '' ⎕EA 'XF ← ⍎ F'
 XF ← f
 T1 ← ( 100 60 60 1000 ) ⊥ ¯4 ↑ ⎕TS
 Z ← Z , T1 - T0
 → ( tr > I ← I + 1 ) / REK
 Z ← ( ÷ ⍴ Z ) × +/ Z
 ⍝DT 12.2.87
∇

⍝ nth Fibonacci number calculation utilizing binomial coefficients.

⍝ F0 ← { (2|⍵) - ⍵ - +/ , M×M = ⍉M ← M ∘.( {⍺ BINOM¨ ⍵} ) ⌽ M ← ¯1 + ⍳⍵ }

⍝ F1 ← { M +.( { ⍺ BINOM¨ ⍵ } ) ( ⍵ - 1 + M ← ¯1 + ⍳ ⌈ ⍵ ÷ 2 ) }

⍝ F0 71

⍝ F1 71

⍝ F0 92

⍝ F1 92

⍝ FF ← { ⍵ , F , ALLFACTORS F ← F1 ⍵ }

⍝ ( ( ⍪ N ) , F1¨ N ) , FF¨ N ← PRIMESTO 60
⍝ ⍪ FF¨ , 53

⍝ )OFF

XF , 1000 TIME '10 BINOM 20'
XF , 1000 TIME '10 ! 20'

XF , 1000 TIME '15 BINOM 30'
XF , 1000 TIME '15 ! 30'

XF , 1000 TIME '15 BINOM 31'
XF , 1000 TIME '15 ! 31'

XF , 1000 TIME '16 BINOM 32'
XF , 1000 TIME '16 ! 32'

XF , 1000 TIME '16 BINOM 33'
XF , 1000 TIME '16 ! 33'

XF , 1000 TIME '17 BINOM 34'
XF , 1000 TIME '17 ! 34'

XF , 1000 TIME '17 BINOM 35'
XF , 1000 TIME '17 ! 35'

XF , 1000 TIME '18 BINOM 36'
XF , 1000 TIME '18 ! 36'

XF , 1000 TIME '18 BINOM 37'
XF , 1000 TIME '18 ! 37'

XF , 1000 TIME '19 BINOM 38'
XF , 1000 TIME '19 ! 38'

XF , 1000 TIME '19 BINOM 39'
XF , 1000 TIME '19 ! 39'

XF , 1000 TIME '20 BINOM 40'
XF , 1000 TIME '20 ! 40'

XF , 1000 TIME '25 BINOM 50'
XF , 1000 TIME '25 ! 50'

XF , 1000 TIME '30 BINOM 60'
XF , 1000 TIME '30 ! 60'

XF , 1000 TIME '30 BINOM 61'
XF , 1000 TIME '30 ! 61'

XF , 1000 TIME '31 BINOM 62'
XF , 1000 TIME '31 ! 62'

XF , 1000 TIME '31 BINOM 63'
XF , 1000 TIME '31 ! 63'

XF , 1000 TIME '32 BINOM 64'
XF , 1000 TIME '32 ! 64'

XF , 1000 TIME '32 BINOM 65'
XF , 1000 TIME '32 ! 65'

XF , 1000 TIME '33 BINOM 66'
XF , 1000 TIME '33 ! 66'

XF , 1000 TIME '34 BINOM 67'
XF , 1000 TIME '34 ! 67'

XF , 1000 TIME '34 BINOM 68'
XF , 1000 TIME '34 ! 68'
